#include <malloc.h>
#include "stdtypes.h"

/****************************************************************************
 Objectives : getting a 8 x 8 block from the original image frame
 Input      : image frame type, a pointer to the original image frame
              the position of the block to be read in the origianl image
 Output     : a 8 X 8 block
 Called By  : muc_coding()
 Calls      : none
****************************************************************************/
long *get_block(u, v, frame_type, image, block) 
int u, v, frame_type;
unsigned char *image;
long block[][8];
{
	unsigned char 	*pointer;		/* pointer to image frame */
	int	x, y;				/* index in a 8 X 8 block */
	int	x_size = MAX_XSIZE;  /* the size of x axis of image frame */
	int	y_size = MAX_YSIZE;  /* the size of y axis of image frame */

	u = (u<<3);
	v = (v<<3);
	for (y=0; y<8; y++) {
		if(frame_type == UFRAME) 
		   pointer = image + (v+y)*(x_size/2+1) + u;
		else if(frame_type == VFRAME) 
		   pointer = image + (v+y)*(x_size/2+1) + u;
		else  /* YFRAME */
		   pointer = image + (v+y)*x_size + u;
		for (x=0; x<8; x++){
			block[y][x] = (long)(*pointer++);
			/* level shift */
			block[y][x] = block[y][x] - 128; 
	 	}
	}
	return (long *)block;
};


/****************************************************************************
 Objectives : putting a encoded/decoded 8 x 8 block to the space for
              a decode image frme data 
 Input      : image frame type, a pointer to the decoded image frame
              the position of the block to be witten to the decoded image
 Output     : none
 Called By  : muc_coding()
 Calls      : range_check()
****************************************************************************/
void put_block(image, b, u, v, frame_type) 
int    u, v, frame_type;
unsigned  char *image;
long  b[][8];
{
	unsigned char 	*pointer;   /* pointer to the decoded image frame */
	long		tmp[8][8]; 		       /* temporary block */	
	long            value;     		       /* temporary value */
	int 		x_size = MAX_XSIZE;/* the size of x-axis of frame */
        int		 x, y;             			 /* index */

	u <<= 3;
	v <<= 3;
	for (y=0; y<8; y++) {
		if(frame_type == UFRAME) 
		   pointer = image + (v+y)*(x_size/2 + 1) + u;
		else if(frame_type == VFRAME) 
		   pointer = image + (v+y)*(x_size/2 + 1) + u;
		else  /* YFRAME */
		   pointer = image + (v+y)*x_size + u;
		for (x=0; x<8; x++) {
			value = range_check(b[y][x], -128, 127);
			*pointer++ = (unsigned char)value + 128;
		}
	}
};

