#include 	<stdio.h> 
#include 	<string.h>
#include        <time.h>
#include        <sys/time.h>
#include	"stdtypes.h"

int NODE;
int tids[32];

int clock_get()
{
        struct timeval tv1;

        gettimeofday(&tv1, (struct timezone*)0);
	return(tv1.tv_sec * 1000000 + tv1.tv_usec);
}

/****************************************************************************
 Objectives : allocate the space for image frames
	      (frame size) Y:U:V = 4:1:1
	      initialize quantization factor  		 	    
	      initialize drawing environment
 Input      : image frame size
 Output     : quantization factor
 Called By  : main()
 Calls      : start_X()
****************************************************************************/
int init(x, y)
int  x, y;
{
	int	size, quant = QUANT; 
  	int	row, col;

	size = sizeof(unsigned char);
	col = x/2 + 1;
	row = y/2;

	/* frame: data structure for the original image 	 */
	/* decoded: data structure for the encoded/decoded image */
	/* initialize Y, U, and V frames */
	frame.y = (unsigned char*)malloc(x * y * size);
	frame.u = (unsigned char*)malloc(col * row * size);
	frame.v = (unsigned char*)malloc(col * row * size);
	decoded.y = (unsigned char*)malloc(x * y * size);
	decoded.u = (unsigned char*)malloc(col * row * size);
	decoded.v = (unsigned char*)malloc(col * row * size);
/*
	if(start_X(x, y) != 0){
	   printf("Xserver failed \n");
	   exit(1);
	}
*/
	return quant;
};

/****************************************************************************
 Objectives : read image frame(Y, U, and V frames)
              from input file to the space allocated
 Input      : image frame data
 Output     : NONE
 Called By  : main()
 Calls      : none
****************************************************************************/
void read_image(fp, frame_type, x_size, y_size) 
FILE *fp;
int  frame_type, x_size, y_size;
{
	int		row, width, height, idx;
	unsigned char	*ptr;

	if(frame_type == YFRAME){
		row = height = y_size;
		width = x_size; 
		ptr = frame.y;
	}
	else if(frame_type == UFRAME){
		row = height = y_size/2;
		width = x_size/2+1; 
		ptr = frame.u;
	}
	else if(frame_type == VFRAME){
		row = height = y_size/2;
		width = x_size/2+1; 
		ptr = frame.v;
	}

	for (idx=0; idx<row*width; idx++) 
		ptr[idx] = fgetc(fp);
	printf("READ IMAGE, Frame Type = %d\t Size = %d\n", frame_type, idx);
};

/****************************************************************************
 Objectives : read the part of image frames, which will be processed in master
 Input      : image frame data, size of the part of the frames
 Output     : NONE
 Called By  : main()
 Calls      : none
****************************************************************************/
void make_local_image(frame_type, line, mline, x, image)
int frame_type, line, mline, x;
unsigned  char  *image;
{
	unsigned char  *ptr;    /* pointer to image frame */
	unsigned char  *dst;    /* pointer to part of image */
	int offset;		/* offset in a image file */
	int mheight;		/* no. of pixels in y axis in master */
	int height;		/* no. of pixels in y axis in each slave */
	int width;		/* no. of pixels in x axis in each slave */
	int size = sizeof(unsigned char);
	int index;

	dst = image;
	switch(frame_type){
	  case 0: /* Y frame */
 		 mheight = mline * 16;
 		 height = line * 16;
		 width = x;
		 ptr = frame.y; 
		 break;
	  case 1: /* U frame */
 		 mheight = mline * 8;
 		 height = line * 8;
		 width = x/2 + 1;
		 ptr = frame.u; 
		 break;
	  case 2: /* V frame */
 		 mheight = mline * 8;
 		 height = line * 8;
		 width = x/2 + 1;
		 ptr = frame.v; 
		 break;
	};
	offset = height*width*(NODE-1);
	ptr = ptr+offset; 

	/* copy some parts of a image */
	for(index = 0; index < width*mheight; index++){
		dst[index] = ptr[index];
	}
}

/****************************************************************************
 Objectives : distribute the part of image frames to the slaves
 Input      : image frame type to be distributed, 
              size of the part of the frames
 Output     : NONE
 Called By  : main()
 Calls      : none
****************************************************************************/
void distribute_image(frame_type, line, x)
int frame_type, line, x;
{
	int msg_type;		/* message type */
	unsigned char  *ptr;    /* pointer to image frame */
	int to;			/* destination node */
	int offset;		/* offset in a image file */
	int data[2];            /* data[0]: heights, data[1]: width */
	int size = sizeof(unsigned char);
	int size1 = 2;

	switch(frame_type){
	  case 0: /* Y frame */
		  ptr = frame.y; 
 		  data[0] = line * 16;
		  data[1] = x;
		  msg_type = 11;     /* sending Y frame */
		  break;
	  case 1: /* U frame */
		  ptr = frame.u; 
 		  data[0] = line * 8;
		  data[1] = x/2 + 1;
		  msg_type = 21;     /* sending U frame */
		  break;
	  case 2: /* V frame */
		  ptr = frame.v; 
 		  data[0] = line * 8;
		  data[1] = x/2 + 1;
		  msg_type = 31;     /* sending V frame */
		  break;
	};

	for(to=0; to<NODE; to++){
	   /* send information on frame size */
	   pvm_initsend(1);
	   pvm_pkint(data, size1, 1);
	   pvm_send(tids[to], msg_type);
	   /* send frame data */
	   offset = data[0] * to * data[1];
	   pvm_initsend(1);
	   pvm_pkbyte(ptr+offset, data[1]*data[0]*size, 1);
	   pvm_send(tids[to], msg_type);
	}
}

/****************************************************************************
 Objectives : read the result of codec from master
 Input      : space to store the result frames
              size of the part of the frames
 Output     : NONE
 Called By  : main()
 Calls      : none
****************************************************************************/
void read_from_master(line, mline, col, prev)
int  line, mline, col;
FRAME  *prev;
{
	int  width, height, i, j, offset, row;
	unsigned  char  *ptr, *tmp;
	FILE *fp;

	/* YFRAME */
	ptr = prev->y;
	tmp = decoded.y;
	height = line * 16;
	offset =  height*col*(NODE-1);
	tmp = tmp + offset;
	row = mline * 16;
	for(i=0; i<row*col; i++)
	      *tmp++ = ptr[i];

	/* UFRAME */
	ptr = prev->u;
	tmp = decoded.u;
	height = line * 3;
	offset =  height*(NODE-1);
	tmp = tmp + offset;
	width = col/2+1;
	row = mline * 8;
	for(i=0; i<row*width; i++)
	      tmp[i] = ptr[i];

	/* VFRAME */
	ptr = prev->v;
	tmp = decoded.v;
	tmp = tmp + offset;
	for(i=0; i<row*width; i++)
	      tmp[i] = ptr[i];
}

/****************************************************************************
 Objectives : read the result of codec sent by the slaves
 Input      : frame type, slave id, image frames sent by the slaves
              size of the part of the frames
 Output     : NONE
 Called By  : receive_result()
 Calls      : none
****************************************************************************/
void read_decoded_frame(frame_type, proc, buff, row, col)
int  frame_type, proc, row, col;
unsigned  char  *buff;
{
	unsigned  char  *ptr;   /* pointer to frame */
	int offset;		/* offset on the decoded frame */
	int i, j;


	switch(frame_type){
	  case YFRAME: ptr = decoded.y; break;
	  case UFRAME: ptr = decoded.u; break;
	  case VFRAME: ptr = decoded.v; break;
	}

	offset = (proc-1)*row*col;
	ptr = ptr + offset;

	for(i=0; i<row*col; i++)
	     *ptr++ = *buff++;
}

/****************************************************************************
 Objectives : receive the result of codec from the slaves
 Input      : message size 
 Output     : NONE
 Called By  : main()
 Calls      :  read_decoded_frame()
****************************************************************************/
void receive_result(line, col, total_rx)
int  line, col, *total_rx;
{
  	int msg_type;		/* type of message 		*/
	int size;		/* size of message 		*/
	unsigned char buffer[150000];  /* pointer to a message buffer  */
	int from;		/* slave process 		*/
	int row;		/* height of the received frame */
	int init_com, end_com;  /* start and end of communicat. */

	for(from=0; from<NODE; from++){
	   /* YFRAME */
	   msg_type = 12;
	   row = line * 16;
           size = row * col * sizeof(unsigned char);
	   init_com = clock_get();
	   pvm_recv(tids[from], msg_type);
	   pvm_upkbyte(buffer, size, 1);
	   end_com = clock_get();
	   *total_rx = *total_rx + end_com - init_com; 
	   printf("RECEIVE RESULT %d\n", from);
	   /* read the part of decoded image frame */
	   read_decoded_frame(YFRAME, from, buffer, row, col);

	   /* UFRAME */
	   msg_type = 22;
	   row = line * 8;
           size = row * (col/2+1) * sizeof(unsigned char);
	   init_com = clock_get();
	   pvm_recv(tids[from], msg_type);
	   pvm_upkbyte(buffer, size, 1);
	   end_com = clock_get();
	   *total_rx = *total_rx + end_com - init_com; 
	   /* read the part of decoded image frame */
	   read_decoded_frame(UFRAME, from, buffer, row, col/2+1);

	   /* VFRAME */
	   msg_type = 32;
	   init_com = clock_get();
	   pvm_recv(tids[from], msg_type);
	   pvm_upkbyte(buffer, size, 1);
	   end_com = clock_get();
	   *total_rx = *total_rx + end_com - init_com; 
	   /* read the part of decoded image frame */
	   read_decoded_frame(VFRAME, from, buffer, row, col/2+1);
	}
}

/****************************************************************************
 Objectives : main routine of jpeg simulation program
 Input      : NONE
 Output     : NONE
 Called By  : System
 Calls      :  read_image(), distribute_image(), make_local_imag(), encode()
	       read_from_master(), receive_result(), 
****************************************************************************/

main(argc, argv)
int argc;
char **argv;
{
	FILE 	*fp, *fp1;
	int	quant;  	/* quantization factor   */
        int     x_size, y_size; /* input image file size */
        int     row, col;	/* image frame size      */
	int 	frame_type;	/* frame type 		 */
	int	mcu_per_line;   /* no. of mcu(minimum coded unit) per line   */
        int     total_line;     /* total no. of mcu lines             	     */
	int 	line;		/* no. of MCU lines being processed in slave */
	int 	mline;		/* no. of MCU lines being processed in master*/
	FRAME   part_of_frame; 	/* part of image frame 			     */
	FRAME	prev;		/* decoded image frame for the part of image */
	char	image_file_name[20];  /* input file name                     */
	int	start, end;           /* starting time, ending time          */
	int     init_time, codec_time;/* elapsed time to initialize and codec*/
	int     init_com, end_com;    /* start and end of communication      */
	int     total_tx; /* total elapsed time to distribute data to each node       */
	int     total_rx; /* total elapsed time to receive the results from each node */
	int     i;


/* ADDED by S.Y. Park : Feb 12, 1994 => Send total ids */

        fp=fopen("no_nodes","r");
        fscanf(fp,"%d",&NODE);
        fclose(fp);
	printf("\nNumber of nodes is %d.\n", NODE);

        pvm_spawn("jpeg_slave", (char **)0, 0, "", NODE, tids);
        for ( i = 0 ; i < NODE ; ++i ) {
            printf("Node [%d] has Tid =  %x\n", i, tids[i]);
            pvm_initsend(1);
            pvm_pkint(&NODE, 1, 1);
            pvm_pkint(tids, NODE, 1);
            pvm_send(tids[i], 0); 
	}
/* END */

	x_size = MAX_XSIZE;
	y_size = MAX_YSIZE;
	mcu_per_line = x_size >> 4;
	total_line = y_size >> 4;

	line = total_line / NODE;
	mline = total_line - line*(NODE-1);

	printf("JPEG image compression demo\n");
	/* init memory for an image */
	quant = init(x_size, y_size);  

	/* init of memory for the local frame */
	row = mline * 16;
	col = x_size;
	part_of_frame.y = (unsigned char *)malloc(row*col);
	prev.y = (unsigned char *)malloc(row*col);
	row = mline * 8;
	col = x_size/2 + 1;
	part_of_frame.u = (unsigned char *)malloc(row*col);
	part_of_frame.v = (unsigned char *)malloc(row*col);
	prev.u = (unsigned char *)malloc(row*col);
	prev.v = (unsigned char *)malloc(row*col);
	
	/***********************************/
	/* read and distribute image files */
	/***********************************/

	total_tx = 0;
	total_rx = 0;
	start = clock_get();
	/* read a luminance file: Y frame  and distribute */
	if ((fp = fopen("ms_y000.dat", "r")) == NULL){
	 	printf("Can't open ms_b000.dat\n");
		exit(1);
	}
	read_image(fp, YFRAME, x_size, y_size);
	if(NODE > 1){
	  init_com = clock_get();
	  distribute_image(YFRAME, line, x_size);
	  end_com = clock_get();
	  total_tx = total_tx + end_com - init_com; 
	}
	make_local_image(YFRAME, line, mline, x_size, part_of_frame.y);
	fclose(fp);

	/* read a chrominance file: V frame  and distribute */
	if ((fp = fopen("ms_r000.dat", "r")) == NULL){
	 	printf("Can't open ms_b000.dat\n");
		exit(1);
	}
	read_image(fp, VFRAME, x_size, y_size);
	if(NODE > 1){
	  init_com = clock_get();
	  distribute_image(VFRAME, line, x_size);
	  end_com = clock_get();
	  total_tx = total_tx + end_com - init_com; 
	}
	make_local_image(VFRAME, line, mline, x_size, part_of_frame.v);
	fclose(fp);

	/* read a chrominance file: U frame  and distribute */
	if ((fp = fopen("ms_b000.dat", "r")) == NULL){
	 	printf("Can't open ms_b000.dat\n");
		exit(1);
	}
	read_image(fp, UFRAME, x_size, y_size);
	if(NODE > 1){
	  init_com = clock_get();
	  distribute_image(UFRAME, line, x_size);
	  end_com = clock_get();
	  total_tx = total_tx + end_com - init_com; 
	}
	make_local_image(UFRAME, line, mline, x_size, part_of_frame.u);
	fclose(fp);
	end = clock_get();
	init_time = end - start;

	/* draw original image frame */
/* 	draw_image_X(frame.y, 0, 0);	   */

	start = clock_get();
	/* encode the part of image allocated to master */
	printf("MASTER: START CODEC\n");
	encode(quant, mline, mcu_per_line, &part_of_frame, &prev);
	printf("MASTER: END OF CODEC\n");

	/* receive the results from slave processes                           */
	/* the elapsed time to receive the results is measured in the routine */
	if ( NODE > 1 ) {
	  receive_result(line, x_size, &total_rx);
	}

        /* read the decoded image frame allocated to master */
	read_from_master(line, mline, x_size, &prev);

	end =  clock_get();
	codec_time = end - start;

	/* draw decoded image frame */
/* 	draw_image_X(decoded.y, 361, 0); */
	printf("Total number of processors used to simulate JPEG = %d\n", NODE);
	printf("Total elapsed time to read and distribute image data = %f ms \n", init_time/1000.0);
	printf("Total elapsed time to distribute image data to slaves = %f ms \n", total_tx/1000.0);
	printf("Total elapsed time for codec = %f ms \n", codec_time/1000.0);
	printf("Total elapsed time to receive the results from slaves = %f ms \n", total_rx/1000.0);
	printf("Total communication time = %f ms \n", (total_rx + total_tx)/1000.0);
	printf("Total elapsed time for encoding and decoding = %f ms \n", (codec_time+init_time)/1000.0);
	printf("\nPress RETURN to finish! \n");

	getchar();
/*        end_X(); */

	pvm_exit();
	
}  /* end of main */

