#include 	<stdio.h>
#include 	<string.h>
#include 	"stdtypes.h"

int NODE;
int my_id;
int tids[32];

/****************************************************************************
 Objectives : read the frame data from the message buffer
 Input      : size of the part of the frames
 Output     : NONE
 Called By  : slave()
 Calls      : none
****************************************************************************/
void read_part_of_frame(buff, part, row, col)
unsigned  char  *buff, *part;
int       row, col;
{
	int i, j, k = 0;            /* index        */

	printf("NODE %d: READ PART OF IMAGE  ", my_id);

	for(i=0; i<row; i++)
	   for(j=0; j<col; j++){
	      part[k] = buff[k];
		k++;
	   }
	printf("(%d x %d)\n", (i-1), (j-1));
};

/****************************************************************************
 Objectives : main routine of the slave processes
 Input      : NONE
 Output     : NONE
 Called By  : System
 Calls      : read_read_part_of_frame(), encode()
****************************************************************************/

main()
{
	unsigned  char  buffer[150000];	/* pointer to a message buffer      */
	unsigned  char  *tmp;   	/* pointer to a message buffer      */
	int     size_info[10];     	/* information for the frame size   */
	int	height, width, col;     /* data size: height * width        */
	int 	my_id;			/* process id       		    */
	int 	src;			/* source process 		    */
	int     to;			/* destination process 		    */
	int	msg_type;		/* type of message 		    */
	int	size;			/* size of message to be received   */
	int     mcu_per_line;   	/* no. of MCUs in each line 	    */
	int	quant;          	/* quantization factor 		    */
	FRAME   part_of_frame, prev; /* part of image frame to be processed */
	int     line;          /* no. of MCU lines being processed in slave */
	int     i;

/* ADDED by S.Y. Park : Feb 12, 1994 => Get total ids and my id */
        my_id = pvm_mytid();
        pvm_recv(pvm_parent(), 0);
        pvm_upkint(&NODE, 1, 1);
        pvm_upkint(tids, NODE, 1);
        for ( i = 0 ; i < NODE ; ++i ) 
            if ( tids[i] == my_id ) { my_id = i; break; }
        printf("\nOK, Node [%d] is awake.\n", my_id);
/* END */

	src = MASTER;

	/********************************/
	/* receive the data from MASTER */
	/********************************/

	/* YFRAME */
  	msg_type = 11;	
	size = 2;
	/* receive the information for the frame data size */
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkint(size_info, size, 1);
	height = size_info[0];
	width = size_info[1];
	col = width;
	printf("NODE %d received %d x %d Y frame data\n",my_id,height,width);

	/* receive the frame data */
	size = height*width*sizeof(unsigned char);
	line = height >> 4;
	mcu_per_line = width >> 4;
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkbyte(buffer, size, 1);

	/* initialize the data structures for the part of image frames */
	part_of_frame.y = 
		(unsigned char*)malloc(width*height*sizeof(unsigned char));
	part_of_frame.u = 
	    (unsigned char*)malloc((width/2+1)*height*sizeof(unsigned char));
	part_of_frame.v = 
	    (unsigned char*)malloc((width/2+1)*height*sizeof(unsigned char));
	prev.y = (unsigned char*)malloc(width*height*sizeof(unsigned char));
	prev.u = 
           (unsigned char*)malloc((width/2+1)*height*sizeof(unsigned char));
	prev.v = 
           (unsigned char*)malloc((width/2+1)*height*sizeof(unsigned char));

	/* read the part of image frame */
	tmp = part_of_frame.y;
	read_part_of_frame(buffer, tmp, height, width);

	/* UFRAME */
  	msg_type = 21;	
	size = 2;
	/* receive the information for the frame data size */
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkint(size_info, size, 1);
	height = size_info[0];
	width = size_info[1];
	printf("NODE %d received %d x %d U frame data\n",my_id,height,width);

	/* receive the frame data */
	size = height*width*sizeof(unsigned char);
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkbyte(buffer, size, 1);

	/* read the part of image frame */
	read_part_of_frame(buffer, part_of_frame.u, height, width);

	/* VFRAME */
  	msg_type = 31;	
	size = 2;
	/* receive the information for the frame data size */
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkint(size_info, size, 1);
	height = size_info[0];
	width = size_info[1];
	printf("NODE %d received %d x %d V frame data\n",my_id,height,width);

	/* receive the frame data */
	size = height*width*sizeof(unsigned char);
	pvm_recv(pvm_parent(), msg_type);
	pvm_upkbyte(buffer, size, 1);

	/* read the part of image frame */
	read_part_of_frame(buffer, part_of_frame.v, height, width);

	printf("NODE %d: SRAT OF CODEC !!\n", my_id);

	/* encode the part of image allocated to slave */
	encode(quant=QUANT, line, mcu_per_line, &part_of_frame, &prev);

	printf("NODE %d: END OF CODEC !!\n", my_id);

	/**************************************/
	/* send the results to MASTER process */
	/**************************************/

	to = src;
	size = sizeof(unsigned char);

	/* YFRAME */
	msg_type = 12;
	height = line * 16;
	width = col;
	pvm_initsend(1);
	pvm_pkbyte(prev.y, height*width*size, 1);
	pvm_send(pvm_parent(), msg_type);
	printf("NODE %d: END OF SENDING YFRAME !!\n", my_id);

	/* UFRAME */
	msg_type = 22;
	height = line * 8;
	width = col/2+1;
	pvm_initsend(1);
	pvm_pkbyte(prev.u, height*width*size, 1);
	pvm_send(pvm_parent(), msg_type);
	printf("NODE %d: END OF SENDING UFRAME !!\n", my_id);

	/* VFRAME */
	msg_type = 32;
	pvm_initsend(1);
	pvm_pkbyte(prev.v, height*width*size, 1);
	pvm_send(pvm_parent(), msg_type);
	printf("NODE %d: END OF SENDING VFRAME !!\n", my_id);

	pvm_exit();

}  /* end of slave */

